import base64
import hashlib
import logging
import warnings
import jcs

from typing import Tuple
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ed25519, padding, rsa
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
)
import based58

logger = logging.getLogger(__name__)

multicodec_ed25519_public_bytes = b"\xed\x01"
multicodec_ed25519_private_bytes = b"\x80\x26"
multicodec_rsa_public_bytes = b"\x85$"


def content_digest_sha256(content: str | bytes) -> str:
    """Computes the SHA256 digest of given content"""
    if isinstance(content, str):
        content = content.encode("utf-8")

    digest = base64.standard_b64encode(hashlib.sha256(content).digest()).decode("utf-8")
    return "sha-256=" + digest


def sign_message(private_key, message):
    warnings.warn("Deprecated will be removed with bovine 0.6.0", DeprecationWarning)
    try:
        key = load_pem_private_key(private_key.encode("utf-8"), password=None)
        assert isinstance(key, rsa.RSAPrivateKey)
    except Exception as e:
        logger.error(e)
        logger.error(private_key)
        raise (e)

    return base64.standard_b64encode(
        key.sign(
            message.encode("utf-8"),
            padding.PKCS1v15(),
            hashes.SHA256(),
        )
    ).decode("utf-8")


def verify_signature(public_key, message, signature):
    public_key_loaded = load_pem_public_key(public_key.encode("utf-8"))

    assert isinstance(public_key_loaded, rsa.RSAPublicKey)

    try:
        public_key_loaded.verify(
            base64.standard_b64decode(signature),
            message.encode("utf-8"),
            padding.PKCS1v15(),
            hashes.SHA256(),
        )
    except InvalidSignature:
        logger.warning("invalid signature")
        return False

    return True


def public_key_to_did_key(public_key: ed25519.Ed25519PublicKey) -> str:
    public_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw,
    )

    encoded = based58.b58encode(multicodec_ed25519_public_bytes + public_bytes)

    return "did:key:z" + encoded.decode("utf-8")


def private_key_to_base58(private_key: ed25519.Ed25519PrivateKey) -> str:
    private_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PrivateFormat.Raw,
        encryption_algorithm=serialization.NoEncryption(),
    )
    encoded = based58.b58encode(multicodec_ed25519_private_bytes + private_bytes)
    return "z" + encoded.decode("ascii")


def did_key_to_public_key(did: str) -> ed25519.Ed25519PublicKey:
    if not did.startswith("did:key:z"):
        raise ValueError(f"Should start with 'did:key:z' got {str}")
    decoded = based58.b58decode(did[9:].encode("ascii"))
    if decoded[:2] != multicodec_ed25519_public_bytes:
        raise ValueError(f"Improper start for ed25519 public key. Got {str}")
    return ed25519.Ed25519PublicKey.from_public_bytes(decoded[2:])


def private_key_to_ed25519(private_key_str: str) -> ed25519.Ed25519PrivateKey:
    decoded = based58.b58decode(private_key_str[1:].encode("utf-8"))
    if decoded[:2] != multicodec_ed25519_private_bytes:
        raise ValueError(f"Improper start for ed25519 private key. Got {str}")

    return ed25519.Ed25519PrivateKey.from_private_bytes(decoded[2:])


def multibase_58btc_encode(data: bytes) -> str:
    """Encodes `data` in base 58 using the bitcoin alphabet
    and adds the prefix `z`"""
    return "z" + based58.b58encode(data).decode("utf-8")


def multibase_decode(data: str) -> bytes:
    """Decodes the string data using the multibase algorithm"""
    if data[0] == "z":
        return based58.b58decode(data[1:].encode("utf-8"))

    raise ValueError(f"{data} encoded in unknown format")


def jcs_sha356(doc: dict) -> str:
    """Returns the sha256 digest of the representation
    of dict according to JCS. This assumes that `doc`
    is JSON serializable.

    JCS is defined in [RFC8785](https://www.rfc-editor.org/rfc/rfc8785).
    """
    return hashlib.sha256(jcs.canonicalize(doc)).digest()


def split_according_to_fep8b32(doc: dict) -> Tuple[dict, dict, bytes]:
    pure_doc = {key: value for key, value in doc.items() if key != "proof"}
    pure_proof = {
        key: value for key, value in doc["proof"].items() if key != "proofValue"
    }
    signature = multibase_decode(doc["proof"]["proofValue"])

    return pure_doc, pure_proof, signature



import pytest

from .types import CryptographicIdentifier
from .test import public_key, did_key


@pytest.mark.parametrize(
    "example",
    [
        {},
        {"type": "Key"},
        {"type": "Multikey"},
        {"type": "Multikey", "controller": "controller"},
        {"type": "Multikey", "publicKeyMultibase": "zxxx"},
        {"type": "Multikey", "controller": "controller", "publicKeyMultibase": "zxxx"},
    ],
)
def test_cryptographic_identity_from_multikey_error(example):
    with pytest.raises(ValueError):
        CryptographicIdentifier.from_multikey(example)


def test_cryptographic_identity_from_multikey():
    public_key_ed25519 = did_key.removeprefix("did:key:")
    controller = "https://com.example/issuer/123"
    multikey = {
        "id": f"{controller}#key-0",
        "type": "Multikey",
        "controller": controller,
        "publicKeyMultibase": public_key_ed25519,
    }

    identity = CryptographicIdentifier.from_multikey(multikey)

    assert identity.controller == controller

    signature = "0gHDz8Qn4+8gWC/byjTk7yvJxL0p4kiUSdxt6VZQvWQ9MBlRThMBDSrgJsNPHZWNMXtPSoL9+r0k\n3cUwYmIWDA=="

    assert identity.verify("moo", signature)
    assert identity.as_tuple() == (controller, public_key_ed25519)


def test_cryptographic_identity_from_multikey_rsa():
    controller = "https://com.example/issuer/123"
    public_key_rsa = "z4MXj1wBzi9jUstyQ1t9N9BHZpoQb4FQpaUfcc61XQigH9ua5R9aEN7YMK81PTKgWLGZXMqvy9eP4X42KgEsmNnZrwQbeT5R8oKhMHVQ5qcwTFwdX2gQeZLLDkUDJL3aJqmqSLK7mrgZ1CskMyD4p8eqFEUW1oufy9cE6Wyz7TQZFKpSCd1oY8HNue9cNRthZzXCdoX6DGVyewBFdivkohE1mhU1EpbKSYH66rx1cZpa6PJKzg4LbKSUqhHaftmsD1jWzFrKNUFzRmCGsihAjLVgsfjPaPmBUXNjYTFg1nCHWCGVGD3g9NhBwGiuu4vQR5PQfD6BCPZpGTaUZjWgZTHveef1pUDPvsCRuGDoGvrTnG8k7SeQp"
    multikey = {
        "id": f"{controller}#key-0",
        "type": "Multikey",
        "controller": controller,
        "publicKeyMultibase": public_key_rsa,
    }

    identity = CryptographicIdentifier.from_multikey(multikey)

    assert identity.controller == controller
    assert identity.as_tuple()[0] == controller

    identity_too = CryptographicIdentifier.from_tuple(*identity.as_tuple())

    assert identity.public_key == identity_too.public_key


@pytest.mark.parametrize(
    "example",
    [
        {},
        {"owner": "owner"},
        {"publicKeyPem": "xxxx"},
        {"owner": "owner", "publicKeyPem": "xxxx"},
    ],
)
def test_cryptographic_identity_from_public_key_error(example):
    with pytest.raises(ValueError):
        CryptographicIdentifier.from_public_key(example)


def test_cryptographic_identity_from_public_key():
    controller = "https://com.example/issuer/123"

    public_key_dict = {
        "id": f"{controller}/main-key",
        "owner": controller,
        "publicKeyPem": public_key,
    }

    identity = CryptographicIdentifier.from_public_key(public_key_dict)

    assert identity.controller == controller

    signature = "vaSYmwpEhGhB/o5QNC8MxYbJeBKDDiaZG0J4EsN0V/5+bFRgPbFK1oUgrdkiTT+farWXdVagPvIg44M/IYjPY8oExBS3mCt9oXDDWiDfBED8n2yrGHV6X/GWNxWUarmo4RcOBU2xWy9982/ZH+UyiPVEpanPi4REf9UYiF0dciZK1Yx3Nkqadnm9XTJJISHX4v88jkUGYaNnWcJ+SJMXuqklYJU/j8j4FVvf3vbFvuwGX1W5o7Zmk89xdJeRiGPYCM2zUgfzGoDHdVHuX8ksR+/xjzwLxMn/SerHuKSCzYivCpaxUqmX0VMTEwvPZ2H+hvXsqyLgR+zFnL7WdX6p7Q=="

    assert identity.verify("secret", signature)
    assert not identity.verify("secret", "")


@pytest.mark.parametrize(
    "example",
    ["", "zxxxxx", "did:key:zxdsafh"],
)
def test_cryptographic_identity_from_did_key_error(example):
    with pytest.raises(ValueError):
        CryptographicIdentifier.from_did_key(example)


def test_cryptographic_identity_from_did_key():
    identifier = CryptographicIdentifier.from_did_key(did_key)

    assert identifier.controller == did_key



import logging
import json
import warnings

from urllib.parse import urlparse
from dataclasses import dataclass
from typing import Callable, Awaitable, Tuple
import bovine.utils
from bovine.utils import parse_gmt

from .helper import content_digest_sha256
from .http_signature import HttpSignature
from .signature import parse_signature_header
from .types import CryptographicIdentifier

logger = logging.getLogger(__name__)


@dataclass
class SignatureChecker:
    """Dataclass to encapsulate the logic of checking a HTTP signature

    :param key_retriever: used to resolve the keyId to the cryptographic information"""

    key_retriever: Callable[
        [str], Awaitable[Tuple[str | None, str | None] | CryptographicIdentifier | None]
    ]

    async def validate_signature(
        self,
        method: str,
        url: str,
        headers: dict,
        body: Callable[[], Awaitable[str | bytes]],
    ) -> str | None:
        """Valids a given signature

        :param method: The http method either get or post
        :param url: The url being queried
        :param headers: The request headers
        :param body: A coroutine resolving the the request body. Used for post requests to check the digest.
        """
        if "signature" not in headers:
            logger.debug("Signature not present on request for %s", url)
            logger.debug(json.dumps(dict(headers)))
            return None

        if method.lower() == "post":
            digest = content_digest_sha256(await body())
        else:
            digest = None

        if digest is not None:
            request_digest = headers["digest"]
            request_digest = request_digest[:4].lower() + request_digest[4:]
            if request_digest != digest:
                logger.warning("Different digest")
                return None

        try:
            http_signature = HttpSignature()
            parsed_signature = parse_signature_header(headers["signature"])
            signature_fields = parsed_signature.fields

            if (
                "(request-target)" not in signature_fields
                or "date" not in signature_fields
            ):
                logger.warning("Required field not present in signature")
                return None

            if digest is not None and "digest" not in signature_fields:
                logger.warning("Digest not present, but computable")
                return None

            http_date = parse_gmt(headers["date"])
            if not bovine.utils.check_max_offset_now(http_date):
                logger.warning(f"Encountered invalid http date {headers['date']}")
                return None

            for field in signature_fields:
                if field == "(request-target)":
                    method = method.lower()
                    parsed_url = urlparse(url)
                    path = parsed_url.path
                    http_signature.with_field(field, f"{method} {path}")
                else:
                    http_signature.with_field(field, headers[field])

            key_result = await self.key_retriever(parsed_signature.key_id)

            if isinstance(key_result, tuple):
                warnings.warn(
                    "Returning a tuple from key_retriever is deprecated, return a CryptographicIdentifier instead, will be remove in bovine 0.6.0",
                    DeprecationWarning,
                )
                key_result = CryptographicIdentifier.from_pem(*key_result)

            if key_result is None:
                logger.debug(f"Could not retrieve key from {parsed_signature.key_id}")
                return None

            return http_signature.verify_with_identity(
                key_result, parsed_signature.signature
            )

        except Exception as e:
            logger.exception(str(e))
            logger.error(headers)
            return None

        return None

    async def validate_signature_request(self, request) -> str | None:
        """Valids a given signature

        :param request: The request object"""
        return await self.validate_signature(
            request.method, request.url, request.headers, request.get_data
        )



import base64
import logging

from dataclasses import dataclass
from cryptography.exceptions import InvalidSignature

from cryptography.hazmat.primitives.asymmetric import ed25519, rsa, padding
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.serialization import (
    load_der_public_key,
    load_pem_public_key,
    load_pem_private_key,
)
from typing import Union, Tuple

import bovine.utils

from .helper import (
    private_key_to_ed25519,
    multibase_58btc_encode,
    multibase_decode,
    split_according_to_fep8b32,
    jcs_sha356,
    multicodec_ed25519_public_bytes,
    multicodec_rsa_public_bytes,
)

logger = logging.getLogger(__name__)


@dataclass
class CryptographicSecret:
    """Represents a cryptographic secret. Such a secret is composed
    of the private key and the URI that resolves to the material, one
    can construct the appropriate cryptographic identifier from.

    :param key_id: The URI, where the corresponding public key and controller
        can be retrieved
    :param private_key: The signing material
    """

    key_id: str
    private_key: Union[ed25519.Ed25519PrivateKey, rsa.RSAPrivateKey]

    def sign(self, message: str):
        """Signs the message.

        Currently only implemented for RSA: Uses PKCS1v15 padding and SHA256
        hashes. Returns the signature as base64 encoded.

        **Warning**: Interface might change, to enable specifying encoding
        of the signature.

        :param message: The message to sign as UTF-8 encoded string."""

        if isinstance(message, str):
            message = message.encode("utf-8")

        if isinstance(self.private_key, rsa.RSAPrivateKey):
            return base64.standard_b64encode(
                self.private_key.sign(
                    message,
                    padding.PKCS1v15(),
                    hashes.SHA256(),
                )
            ).decode("utf-8")
        if isinstance(self.private_key, ed25519.Ed25519PrivateKey):
            return multibase_58btc_encode(self.private_key.sign(message))

        raise ValueError("Unknown key type in private key")

    def fep_8b32_sign(self, document: dict) -> dict:
        """signs the current document with the algorithm given by FEP-8b32"""

        created = bovine.utils.get_gmt_now()
        proof = {
            "created": created,
            "cryptosuite": "eddsa-jcs-2022",
            "proofPurpose": "assertionMethod",
            "type": "DataIntegrityProof",
            "verificationMethod": self.key_id,
        }
        digest = jcs_sha356(proof) + jcs_sha356(document)
        proof["proofValue"] = self.sign(digest)

        return {**document, "proof": proof}

    @staticmethod
    def from_pem(key_id: str, pem: str):
        """Creates a CryptographicSecret from a PEM encoded private key"""
        return CryptographicSecret(
            key_id, load_pem_private_key(pem.encode("utf-8"), password=None)
        )

    @staticmethod
    def from_multibase(key_id: str, multibase: str):
        """Creates a CryptographicSecret from multibase encoded
        Ed25519 private key and key_id"""
        return CryptographicSecret(key_id, private_key_to_ed25519(multibase))


@dataclass
class CryptographicIdentifier:
    """Represents a cryptographic identifier. The usage is: If an object is
    signed by `public_key`, then it is authored by `controller`. In order
    to discover which `CryptographicIdentifier` to use, one resolves another
    identifier `key_id`, which yields either a Multikey or a publicKey object, which
    can then be resolved into a CryptographicIdentifier.

    One should never need to directly access the properties of this class, instead
    verify returns the controller, if and only if the signature is valid.

    :param controller: The URI of the actor that controls the public key
    :param public_key: Public key used to verify signatures
    """

    controller: str
    public_key: Union[ed25519.Ed25519PublicKey, rsa.RSAPublicKey]

    def verify(self, message: str | bytes, signature: str | bytes) -> str | None:
        """Verifies that `signature` is a correct signature for the given message.

        **Warning**: Interface might change, to enable specifying encoding
        of the signature.

        :param message: The message string.
        :param signature: The signature

        :return: If the signature is valid the corresponding controller,
            otherwise null.
        """
        # Doing the encoding here is probably wrong ...
        # All these things are awkward

        if isinstance(signature, str):
            signature = base64.standard_b64decode(signature)

        if isinstance(message, str):
            message = message.encode("utf-8")

        if isinstance(self.public_key, rsa.RSAPublicKey):
            try:
                self.public_key.verify(
                    signature,
                    message,
                    padding.PKCS1v15(),
                    hashes.SHA256(),
                )
                return self.controller
            except InvalidSignature:
                return None

        if isinstance(self.public_key, ed25519.Ed25519PublicKey):
            try:
                self.public_key.verify(signature, message)
                return self.controller
            except InvalidSignature:
                return None

        raise ValueError("Unknown key type in public_key")

    def fep_8b32_verify(self, document: dict):
        """Verifies that document has a valid signature according to FEP-8b32.
        We note that in order to verify a document signed using FEP-8b32, one
        will already need to parse it sufficiently to extract the controller,
        so the CryptographicIdentifier can be created.

        :param document: The document to verify"""
        pure_doc, pure_proof, signature = split_according_to_fep8b32(document)
        digest = jcs_sha356(pure_proof) + jcs_sha356(pure_doc)

        return self.verify(digest, signature)

    def as_tuple(self) -> Tuple[str, str]:
        """Transforms the CryptographicIdentifier into a tuple

        :return: controller, multibase/multicodec encoded public key"""
        if isinstance(self.public_key, rsa.RSAPublicKey):
            public_bytes = multicodec_rsa_public_bytes + self.public_key.public_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
            public_key = multibase_58btc_encode(public_bytes)
        else:
            public_key = multibase_58btc_encode(
                multicodec_ed25519_public_bytes
                + self.public_key.public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw,
                )
            )
        return (self.controller, public_key)

    @staticmethod
    def from_pem(public_key: str, owner: str):
        """Creates a CryptographicIdentifier from a pem encoded public key and the controller"""
        if public_key is None:
            return None
        return CryptographicIdentifier(
            controller=owner, public_key=load_pem_public_key(public_key.encode("utf-8"))
        )

    @staticmethod
    def from_multikey(multikey: dict):
        """Creates a CryptographicIdentifier from a Multikey, see

        * [FEP-521a: Representing actor's public keys](https://codeberg.org/fediverse/fep/src/branch/main/fep/521a/fep-521a.md)

        Example:

        ```json
        {
            "id": "https://server.example/users/alice#ed25519-key",
            "type": "Multikey",
            "controller": "https://server.example/users/alice",
            "publicKeyMultibase": "z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2"
        }
        ```
        """
        if multikey.get("type") != "Multikey":
            raise ValueError("Multikeys must have type 'Multikey'")

        controller = multikey.get("controller")
        public_key = multikey.get("publicKeyMultibase")

        if controller is None or public_key is None:
            raise ValueError(
                "Expected parameters controller and publicKeyMultibase to be present when parsing MultiKey"
            )

        return CryptographicIdentifier.from_tuple(controller, public_key)

    @staticmethod
    def from_tuple(controller: str, multibase_public_key: str):
        """Creates a CryptographicIdentifier from a tuple

        :param controller: The controller URI
        :param multibase_public_key: The public key encoded using multibase/multicodex
        """
        public_key = multibase_decode(multibase_public_key)

        if public_key[:2] == multicodec_ed25519_public_bytes:
            return CryptographicIdentifier(
                controller=controller,
                public_key=ed25519.Ed25519PublicKey.from_public_bytes(public_key[2:]),
            )
        if public_key[:2] == multicodec_rsa_public_bytes:
            return CryptographicIdentifier(
                controller=controller,
                public_key=load_der_public_key(public_key[2:]),
            )

        raise ValueError("Unsupported public key format")

    @staticmethod
    def from_did_key(did_key: str):
        """Creates a cryptographic identifier from a did:key
        The controller is then the did:key and the public key
        the encoded public key.

        :param did_key: The did key, e.g. "did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5"
        """

        if did_key.startswith("did:key:"):
            decoded = multibase_decode(did_key[8:])
            if decoded[:2] == multicodec_ed25519_public_bytes:
                return CryptographicIdentifier(
                    controller=did_key,
                    public_key=ed25519.Ed25519PublicKey.from_public_bytes(decoded[2:]),
                )

        raise ValueError("Invalid did key format")

    @staticmethod
    def from_public_key(data: dict):
        """Creates a Cryptographic identifier from a publicKey object, example:

        ```json
        {
            "id": "https://com.example/issuer/123#main-key",
            "owner": "https://com.example/issuer/123",
            "publicKeyPem": "-----BEGIN PUBLIC KEY-----\\n...\\n-----END PUBLIC KEY-----"
        }
        ```
        """
        controller = data.get("owner")
        public_key = data.get("publicKeyPem")

        if controller is None or public_key is None:
            raise ValueError(
                "Expected parameters owner and publicKeyPem to be present when parsing publicKey"
            )

        return CryptographicIdentifier(
            controller=controller,
            public_key=load_pem_public_key(public_key.encode("utf-8")),
        )



import aiohttp
import bovine.clients.signed_http_methods

from bovine.crypto.types import CryptographicSecret


class SignedHttpClient:
    """Client for using HTTP Signatures"""

    def __init__(self, session, public_key_url, private_key):
        """init

        :param session: The aiohttp.ClientSession
        :param public_key_url: Used as keyId when signing
        :param private_key: The pem encoded private key
        """
        self.session = session
        self.secret = CryptographicSecret.from_pem(public_key_url, private_key)

    async def get(self, url, headers={}) -> aiohttp.ClientResponse:
        """Retrieves url using a signed get request"""
        return await bovine.clients.signed_http_methods.signed_get(
            self.session, self.secret, url, headers
        )

    async def post(self, url, body, headers={}, content_type=None):
        """Posts to url using a signed post request"""
        return await bovine.clients.signed_http_methods.signed_post(
            self.session,
            self.secret,
            url,
            body,
            headers,
            content_type=content_type,
        )

    def event_source(self, url):
        """Opens an event source to url"""
        return bovine.clients.signed_http_methods.signed_event_source(
            self.session, self.secret, url
        )
